package dao

import (
	"context"
	"contract_backend/config"
	"fmt"
	"math/big"
	"os"
	"sync"
	"time"

	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ethereum/go-ethereum/ethclient"
	"github.com/ethereum/go-ethereum/rpc"
	"gorm.io/driver/postgres"
	"gorm.io/gorm"
	"gorm.io/gorm/logger"
	"gorm.io/gorm/schema"
)

type Dao struct {
	chainId        *big.Int
	c              *config.Config
	db             *gorm.DB
	ethClient      *ethclient.Client
	ethRPC         *rpc.Client
	gasSenderNonce uint64
	gasSenderLock  sync.Mutex
	gasPrice       *big.Int
}

func New(_c *config.Config) (dao *Dao, err error) {
	dao = &Dao{
		c: _c,
	}
	dao.ethClient, err = ethclient.Dial(_c.Chain.RPC)
	if err != nil {
		return
	}
	dao.ethRPC, err = rpc.Dial(_c.Chain.RPC)
	if err != nil {
		return
	}
	dsn := fmt.Sprintf("host=%s user=%s password=%s dbname=%s port=%d",
		_c.PGSQL.Host, _c.PGSQL.User, _c.PGSQL.Password, _c.PGSQL.Database, _c.PGSQL.Port,
	)

	chainId, err := dao.ethClient.ChainID(context.Background())
	if err != nil {
		return
	}

	dao.chainId = chainId

	k, _ := crypto.HexToECDSA(_c.Chain.GasSenderPrivateKey)
	dao.gasSenderNonce, _ = dao.ethClient.NonceAt(context.Background(), crypto.PubkeyToAddress(k.PublicKey), nil)
	go dao.LoopGasPrice()

	if _c.PGSQL.CertFile != "" {
		dsn = fmt.Sprintf("%s sslmode=require sslrootcert=%s", dsn, _c.PGSQL.CertFile)
	}

	lgr := logger.Default
	if _c.PGSQL.EnableLog {
		lgr = logger.Default.LogMode(logger.Info)
	}

	dao.db, err = gorm.Open(postgres.Open(dsn), &gorm.Config{
		NamingStrategy: schema.NamingStrategy{
			SingularTable: true,
		},
		Logger: lgr,
	})
	if err != nil {
		return
	}
	sqlDB, err := dao.db.DB()
	if err != nil {
		return
	}
	sqlDB.SetMaxOpenConns(_c.PGSQL.MaxConn)
	sqlDB.SetMaxIdleConns(_c.PGSQL.MaxIdleConn)
	sqlDB.SetConnMaxIdleTime(time.Hour)

	migrate := os.Getenv("MIGRATE")
	if migrate == "true" {
		// err = dao.db.AutoMigrate(&dbModel.User{}, &dbModel.Active{}, &dbModel.ChatGroup{})
		if err != nil {
			return
		}
	}

	return dao, nil
}
