package dao

import (
	"context"
	"math/big"
	"time"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/core/types"
	"github.com/ethereum/go-ethereum/crypto"
	log "github.com/sirupsen/logrus"
)

func (d *Dao) LoopGasPrice() {
	gasPrice, err := d.ethClient.SuggestGasPrice(context.Background())
	if err != nil {
		log.WithError(err).Error("get gas price error")
		panic(err)
	}
	d.gasPrice = big.NewInt(0).Div(
		big.NewInt(0).Mul(gasPrice, big.NewInt(11)),
		big.NewInt(10),
	)

	ticker := time.NewTicker(time.Second * 10)
	defer ticker.Stop()
	for {
		select {
		case <-ticker.C:
			gasPrice, err := d.ethClient.SuggestGasPrice(context.Background())
			if err != nil {
				log.WithError(err).Error("get gas price error")
				continue
			}
			d.gasPrice = big.NewInt(0).Div(
				big.NewInt(0).Mul(gasPrice, big.NewInt(11)),
				big.NewInt(10),
			)
		}
	}
}

// EstimateGasRecharge 估算充值金额
func (d *Dao) EstimateGasRecharge(address common.Address) (val *big.Int, err error) {
	_10val := new(big.Int).Mul(d.gasPrice, big.NewInt(12*500000)) // 10次 每次50w
	_20val := new(big.Int).Mul(d.gasPrice, big.NewInt(24*500000)) // 20次 每次50w

	currentBalance, err := d.ethClient.BalanceAt(context.Background(), address, nil)
	if err != nil {
		return nil, err
	}

	if currentBalance.Cmp(_10val) < 0 {
		return new(big.Int).Sub(_20val, currentBalance), nil
	}

	return nil, nil
}

func (d *Dao) TransferETH(hexPrivateKey string, address common.Address, value *big.Int) (txHash common.Hash, err error) {
	privateKey, _ := crypto.HexToECDSA(hexPrivateKey)
	d.gasSenderLock.Lock()
	defer func() {
		d.gasSenderNonce++
		d.gasSenderLock.Unlock()
	}()
	innerTx := types.LegacyTx{
		Nonce:    d.gasSenderNonce,
		To:       &address,
		Value:    value,
		Gas:      21000,
		GasPrice: d.gasPrice,
	}
	tx := types.NewTx(&innerTx)
	innerTx.GasPrice = d.gasPrice

	signer := types.NewCancunSigner(d.chainId)
	signedTx, err := types.SignTx(tx, signer, privateKey)
	if err != nil {
		return common.Hash{}, err
	}
	return signedTx.Hash(), d.ethClient.SendTransaction(context.Background(), signedTx)
}

func (d *Dao) WaitForReceipt(hash common.Hash) (receipt *types.Receipt, err error) {
	timeoutCtx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
	defer cancel()

	queryTicker := time.NewTicker(time.Second)
	defer queryTicker.Stop()

	for {
		receipt, err := d.ethClient.TransactionReceipt(timeoutCtx, hash)
		if err == nil {
			return receipt, nil
		}

		select {
		case <-timeoutCtx.Done():
			return nil, timeoutCtx.Err()
		case <-queryTicker.C:
		}
	}
}
