package monitor

import (
	"contract_backend/config"
	. "contract_backend/constant"
	"contract_backend/dao"
	"contract_backend/model"
	"crypto/ecdsa"
	"math/big"
	"time"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
	log "github.com/sirupsen/logrus"
)

type Monitor struct {
	conf       *config.Config
	d          *dao.Dao
	privateKey *ecdsa.PrivateKey
}

func NewMonitor(_conf *config.Config, _dao *dao.Dao) *Monitor {
	k, err := crypto.HexToECDSA(common.Bytes2Hex(common.FromHex(_conf.Chain.ExecutorPrivateKey)))
	if err != nil {
		panic(err)
	}
	return &Monitor{
		conf:       _conf,
		d:          _dao,
		privateKey: k,
	}
}

func (m *Monitor) Start() {
	for {
		tasks, err := m.d.GetUnprocessedTasks(1)
		if err != nil {
			log.WithError(err).Error("get unprocessed tasks error")
			continue
		}
		if len(tasks) == 0 {
			time.Sleep(2 * time.Second)
			continue
		}

		log.WithField("tasks", len(tasks)).Info("get unprocessed tasks")

		for _, task := range tasks {
			m.ProcessOne(task)
		}
	}

}

func (m *Monitor) ProcessOne(task *model.ExecTask) {
	err := m.d.UpdateTask(task.Id, "", TxStatusPending)
	if err != nil {
		log.WithFields(log.Fields{
			"id":  task.Id,
			"err": err.Error(),
		}).Error("update task error")
		return
	}
	log.WithFields(log.Fields{
		"task_id":   task.Id,
		"user_id":   task.UserId,
		"app_key":   task.AppKey,
		"model_ids": task.ModelIdsArray,
	}).Info("execute task")

	val := big.NewInt(0).Mul(big.NewInt(1000000000000000000), big.NewInt(int64(m.conf.Chain.ModelPrice)))
	// txHash, err := m.d.ExecuteTask(m.privateKey, task.Id, task.UserId, task.AppKey, task.ModelIdsArray, big.NewInt(int64(m.conf.Chain.ModelPrice)))
	txHash, err := m.d.ExecuteTask(m.privateKey, task.Id, task.UserId, task.AppKey, task.ModelIdsArray, val)
	if err != nil {
		log.WithFields(log.Fields{
			"id":  task.Id,
			"err": err.Error(),
		}).Error("execute task error")
		err = m.d.UpdateTask(task.Id, "", TxStatusFailed)
		return
	}

	receipt, err := m.d.WaitForReceipt(txHash)
	if err != nil {
		log.WithFields(log.Fields{
			"id":  task.Id,
			"err": err.Error(),
		}).Error("wait for receipt error")
		err = m.d.UpdateTask(task.Id, txHash.String(), TxStatusUnknown)
	}

	if receipt.Status == 1 {
		err = m.d.UpdateTask(task.Id, txHash.String(), TxStatusSuccess)
	} else {
		err = m.d.UpdateTask(task.Id, txHash.String(), TxStatusRevert)
	}
	if err != nil {
		log.WithFields(log.Fields{
			"id":  task.Id,
			"err": err.Error(),
		}).Error("update task error")
		return
	}
}
